#include "neural_mlfn.h"

/*-----------------------------------------------------
   - create and initialize a MLFN neural net
   - allocate memory for synapses and neurons
   - init unused neurons/synapses to 0 (VERY important)
   restrictions: no layer may have more than 16 neurons
                 total number of layers must be < 255
                 making num neurons a multiple of 4 increases efficiency
--------------------------------------------------------*/
mlfn_net *create_mlfn_network(const char num_layers, 
                              const char *num_neurons)
{
   mlfn_net *net;
   int l,n;

   /* create the network */
   DEBUG_MALLOC(mlfn_net,net,16,sizeof(mlfn_net),NULL)
   DEBUG_MALLOC(char,net->num_neurons,16,num_layers,NULL)
   memcpy(net->num_neurons,num_neurons,num_layers);
    
   /* number of  input, hidden and output layers */
   net->num_layers = num_layers;
   
   /* allocate actual layers */
   DEBUG_MALLOC(mlfn_layer,net->layer,16,sizeof(mlfn_layer)*net->num_layers,NULL)

   /* allocate input layer neuron vectors */
   net->layer[0].vects = vectors_needed(net->num_neurons[0]);
   DEBUG_MALLOC(ps2_vu0_fvector,net->layer[0].neuron,16,sizeof(ps2_vu0_fvector)*net->layer[0].vects,NULL)

   /* beware input layer synapse dangling pointer */
   net->layer[0].synapse = NULL;

   /* init all other layers and dynamically allocate synapses/neurons */
   for (l=1; l<net->num_layers; l++)
   {
       /* min number of vectors to hold neurons */
       net->layer[l].vects = vectors_needed(net->num_neurons[l]);
 
       /* allocate synapses */
       DEBUG_MALLOC(ps2_vu0_fvector,net->layer[l].synapse,16,sizeof(ps2_vu0_fvector*)*net->num_neurons[l],NULL)

       for(n=0; n<net->num_neurons[l]; n++)
       {
	  DEBUG_MALLOC(ps2_vu0_fvector,net->layer[l].synapse[n],16,sizeof(ps2_vu0_fvector)*net->layer[l-1].vects,NULL)
       }    

       /* allocate neurons */
       DEBUG_MALLOC(ps2_vu0_fvector,net->layer[l].neuron,16,sizeof(ps2_vu0_fvector)*net->layer[l].vects,NULL)
   }

   return(net);
}

/*--------------------------------------------------------
free a dynamically allocated mlfn neural net - very slowly
----------------------------------------------------------*/
char destroy_mlfn_net(mlfn_net *net)
{
    unsigned char l,n;
    /* dont free unallocated memory */
    if (net)
    {
       DEBUG_FREE((ps2_vu0_fvector *)(net->layer[0].neuron)); 
       /* free synapses and neurons */
       for (l=1; l<net->num_layers; l++)
       {
	  for (n=0; n<net->num_neurons[l]; n++)
            DEBUG_FREE((ps2_vu0_fvector*)net->layer[l].synapse[n]);     
          DEBUG_FREE((ps2_vu0_fvector **)(net->layer[l].synapse));
          DEBUG_FREE((ps2_vu0_fvector *)(net->layer[l].neuron));
       }
       /* layers, net, and avoid dangling pointers */
       DEBUG_FREE(net->layer);
       DEBUG_FREE(net->num_neurons);
       DEBUG_FREE(net);
       net = NULL;
    }
    return(0);
}

/*-----------------------------------------------------
   MLFN forward propagation - present an input and 
   get the net output This needs some _serious_ 
   optimizations and unrolling at this point.  I'll
   rewrite it in assembly for the next release.
------------------------------------------------------*/
int mlfn_fwd_propagate(mlfn_net *net)
{
    unsigned char l;
    unsigned char n;

    /* load hermite coefficients to float registers
       this is ok as long as no other function calls
       are made inside this func... otherwise registers
       may get overwritten or trampled
    */ 
    LOAD_MLFN_HERMITE_COEFF
    
    /* for each layer in the net */
    for (l=1; l<net->num_layers; l++)
    {
        /* initiate the first calculation */
        LOAD_NEURONS(net->layer[l-1].neuron)
        LOAD_SYNAPSES(net->layer[l].synapse[0],net->layer[l-1].vects)
	CALL_MICRO(vu0_mha)

        /* Calc the weighted sum for each neuron in the layer.
           The current neuron's weighted sum (vector microcode)
           and the previous neuron's activation (mips core asm)
           are computed in parallel
	*/
        for (n=1; n<net->num_neurons[l]; n++)
	{
	    LOAD_SYNAPSES(net->layer[l].synapse[n],net->layer[l-1].vects)
	    /* activation approx - is this cast too expensive? */
	    mlfn_calc_activation((float*)(&(net->layer[l].neuron[n-1])));
	    CALL_MICRO(vu0_mha)	    
        }
        /* wait for the last micro routine to finish */
        asm __volatile__ ("vnop");
        mlfn_calc_activation((float*)(&(net->layer[l].neuron[n])));
    }  
    return(0);
}

/*-------------------------------------------------------- 
   Given the number of neurons in a layer, calculate the
   number of vectors needed to hold them.
   Must assume neurons in a layer <= 16 
----------------------------------------------------------*/
int vectors_needed(const char neurons)
{
  /* produce slightly faster code */
  if (neurons > 8)
  {
     return((neurons<12)?3:4);
  }
  else
     return((neurons>4)?2:1);
}

/*--------------------------------------------------------
Load and save a previously trained mlfn neural net
Load can also create a new net based on the input file
Note to self: doesn't verify num_layers < 255 
net->layer[L].synapse[neuron][vec]
---------------------------------------------------------*/
mlfn_net *load_mlfn(char *filename, mlfn_net *net, char create)
{
  FILE *fp;
  int fd;
  unsigned char l;
  unsigned char n;
  unsigned char v;
  /* this is questionable.  Should I use dma here instead? */
  char num_neurons[255];

  if ((fd = open(filename,O_RDONLY)) == -1)
     DBG("cant open mlfn file for reading",NULL)  

  /* read number of layers and neurons per layer */
  read(fd,&n,1); 
  read(fd,num_neurons,n);

  /* if we are creating a new net based on the file */
  if (create)
     net = create_mlfn_network(n,num_neurons);

  else if ((!create) && (!net))
     DBG("attempting to use unallocated mlfn net",NULL)
  
  /* or filling in a pre-allocated net (dangerous! not recommended) */ 
  else
  {
    /* not yet implemented */
  }

  /* read synapses from file.  gcc -O2 unrolls the inner loop */
  if (feof(fp = fdopen(fd,"r")))
    DBG("sudden end of MLFN synapse file",NULL);

  for (l=1; l<net->num_layers; l++)
    for (n=0; n<net->num_neurons[l]; n++)
      for (v=0; v<net->layer[l-1].vects; v++)
        fscanf(fp," %f %f %f %f",&net->layer[l].synapse[n][v][0],
                                 &net->layer[l].synapse[n][v][1],
                                 &net->layer[l].synapse[n][v][2],
                                 &net->layer[l].synapse[n][v][3]); 

       

  /* end of file if there were not enough synapses to read */
  if (feof(fp))
    DBG("not enough synapses in file - net improperly initialized",NULL);

  if (fgetc(fp) != END_SYNAPSE_FILE)
    DBG_WARN("sudden end of mlfn synapse file")

  close(fd);
  return(net);
}

/*-------------------------------------------------------*/
int save_mlfn(char *filename, mlfn_net *net)
{
  FILE *fp;
  int fd;
  unsigned char l;
  unsigned char n; 
  unsigned char v;

  if ((fd = open(filename,O_RDWR|O_CREAT | O_TRUNC,OPEN_ALL_PERM)) == -1)
    DBG("cant open mlfn file for saving",errno)
  
  /* write net info to the file */
  write(fd,&net->num_layers,1);
  write(fd,net->num_neurons,net->num_layers);

  /* get the corresponding pointer from the descriptor */
  if ((fp = fdopen(fd,"w+")) == NULL)  
    DBG("failed to change fdesctiptor to fpointer",errno)

  /* write synapses to file.  gcc -O2 unrolls the inner loop */
  for (l=1; l<net->num_layers; l++)
    for (n=0; n<net->num_neurons[l]; n++)
      for (v=0; v<net->layer[l-1].vects; v++)
        fprintf(fp," %f %f %f %f",net->layer[l].synapse[n][v][0],
                                  net->layer[l].synapse[n][v][1],
                                  net->layer[l].synapse[n][v][2],
                                  net->layer[l].synapse[n][v][3]);
  /* end file and close */
  fprintf(fp,"%c",END_SYNAPSE_FILE);
  fclose(fp);
  close(fd);
  return(0);
}
